#!/bin/bash
# MCM - Machine Configuration Manager; manages the contents of files and directories
# Copyright (c) 2013-2018 Anthony Doggett <mcm@interfaces.org.uk>
# 
# Licence:
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

TEMP=$(mktemp --tmpdir="$TMP" -d 'mcm_runtests_XXXXX')
function cleanup {
    # Perform program exit housekeeping
    EXIT_STATUS=$?
    echo "(Removing all test files from $TEMP)"
    rm -r $TEMP
    trap 0  # reset to default action
    exit $EXIT_STATUS
}
trap cleanup 0 1 2 3 15

if [ $# -eq 0 ]
then
    MCM=$PWD/dist/build/mcm/mcm
    MCM2HTML=$PWD/dist/build/mcm2html/mcm2html
    COMMANDS2HTML=$PWD/dist/build/commands2html/commands2html
elif  [ $# -eq 1 ]
then
    MCM=$1/mcm
    MCM2HTML=$1/mcm2html
    COMMANDS2HTML=$1/commands2html
else
    echo "Expected 0 or 1 argument(s), but received $#"
    exit 1
fi
set -e
test -x $MCM || (echo "Binary $MCM missing"; exit 1)
test -x $MCM2HTML || (echo "Binary $MCM2HTML missing"; exit 1)
test -x $COMMANDS2HTML || (echo "Binary $COMMANDS2HTML missing"; exit 1)
set +e

REGEXTEMP=$(echo $TEMP | sed 's/\//\\\//g')
function diffwithstdin {
    sed "s/$REGEXTEMP/\/tmp/g" | diff $1 -
}

echo "Versions under test:"
echo -n "  MCM: $MCM: "
$MCM -V
echo -n "  MCM2HTML: $MCM2HTML: "
$MCM2HTML -V
echo -n "  COMMANDS2HTML: $COMMANDS2HTML: "
$COMMANDS2HTML -V

echo
echo "Checking expected output when parse errors occur"
set -e
for i in parseerrortests/*.mcm;do
    echo $i
    expected=`echo $i|sed s/mcm/out/g`
    $MCM Test.test -P parseerrortests -f no -r $TEMP/mcm_tests --stdin <$i 2>&1 | diffwithstdin $expected
done

echo
echo "Checking expected output when root does not exist"
set -e
for i in tests/*.mcm;do
    echo $i
    expected=`echo $i|sed s/mcm/out/g`
    $MCM Test.test -P tests -f no -r $TEMP/mcm_tests --stdin <$i | diffwithstdin $expected
done

echo
echo "Checking error output (files do not exist)"
set -e
for i in tests/*.errormcm;do
    echo $i
    expected=`echo $i|sed s/mcm/out/g`
    $MCM Test.test -P tests -f no -r $TEMP/mcm_tests --stdin <$i | diffwithstdin $expected
done

echo
echo "Checking expected output when files exist and are all perfect"
set -e
cp -a testresults/ $TEMP/mcm_tests/
ln -s /tmp/link1 $TEMP/mcm_tests/tmp/createsymlinktest1
ln -s ./link2 $TEMP/mcm_tests/tmp/createsymlinktest2
chgrp cdrom $TEMP/mcm_tests/tmp/permissiontest1
ln -s /tmp/linked $TEMP/mcm_tests/tmp/permissiontest3
chgrp -h cdrom $TEMP/mcm_tests/tmp/permissiontest3
chmod 751 $TEMP/mcm_tests/tmp/permissiontest2
mkdir -p $TEMP/mcm_tests/tmp/createdirtest1
mkdir -p $TEMP/mcm_tests/tmp/createdirtest2
mkdir -p $TEMP/mcm_tests/tmp/createdirtest3
chmod 700 $TEMP/mcm_tests/tmp/createdirtest5
chmod 700 $TEMP/mcm_tests/tmp/createdirtest6
for i in tests/*.mcm;do
    echo $i
    $MCM Test.test -P tests -f no -r $TEMP/mcm_tests --stdin -d <$i | diff /dev/null -
done
echo
echo "Checking expected output when files exist and are all perfect and root is relative"
set -e
cd $TEMP
for i in $OLDPWD/tests/*.mcm;do
    echo $i
    $MCM Test.test -P $OLDPWD/tests -f no -r ./mcm_tests --stdin -d <$i | diff /dev/null -
done
cd -

echo
echo "Checking when root directory is fully manged"
set -e
mkdir -p $TEMP/mcm_manageddir
echo fred >$TEMP/mcm_manageddir/fred
$MCM Test.test -P tests/manageddirtest -f no -r $TEMP/mcm_manageddir --stdin -d <tests/manageddirtest/manageddir.mcm | diff /dev/null -
echo "Checking when root directory is fully manged and root is relative"
cd $TEMP
$MCM Test.test -P $OLDPWD/tests/manageddirtest -f no -r ./mcm_manageddir --stdin -d <$OLDPWD/tests/manageddirtest/manageddir.mcm | diff /dev/null -
cd -

echo
echo "Checking removal of files"
set -e
mkdir -p $TEMP/removaltest
mkdir -p $TEMP/removaltest/emptydir1 -m 777
mkdir -p $TEMP/removaltest/emptydir2 -m 770
mkdir -p $TEMP/removaltest/emptydir3 -m 700
echo fred >$TEMP/removaltest/file1
echo fred >$TEMP/removaltest/file2
chmod 707 $TEMP/removaltest/file2
mkdir -p $TEMP/removaltest/nonemptydir1 -m 777
echo mavis >$TEMP/removaltest/nonemptydir1/mavis1
mkdir -p $TEMP/removaltest/nonemptydir2 -m 770
echo mavis >$TEMP/removaltest/nonemptydir2/mavis2
mkdir -p $TEMP/removaltest/nonemptydir3 -m 707
echo mavis >$TEMP/removaltest/nonemptydir3/mavis3
chmod 707 $TEMP/removaltest/nonemptydir3/mavis3
echo all
$MCM Test.all     -P tests/removaltest -f no -r $TEMP/removaltest --stdin -d <tests/removaltest/removaltest.mcm | diffwithstdin tests/removaltest/all.out
echo single
$MCM Test.singles -P tests/removaltest -f no -r $TEMP/removaltest --stdin -d <tests/removaltest/removaltest.mcm | diffwithstdin tests/removaltest/singles.out
echo "Checking removal of files when root is relative"
cd $TEMP
echo 'all (relative)'
$MCM Test.all     -P $OLDPWD/tests/removaltest -f no -r ./removaltest --stdin -d <$OLDPWD/tests/removaltest/removaltest.mcm | diffwithstdin $OLDPWD/tests/removaltest/relative_all.out
echo 'single (relative)'
$MCM Test.singles -P $OLDPWD/tests/removaltest -f no -r ./removaltest --stdin -d <$OLDPWD/tests/removaltest/removaltest.mcm | diffwithstdin $OLDPWD/tests/removaltest/relative_singles.out
cd -

echo
echo "Checking --fix functionality"
set -e
$MCM -P tests/fixfunctionality Fix.createFile -r $TEMP/fixfunctionality -f no | diffwithstdin tests/fixfunctionality/fix_no.out
echo no | $MCM -P tests/fixfunctionality Fix.createFile -r $TEMP/fixfun_prompt -f prompt | diffwithstdin tests/fixfunctionality/fix_prompt_no.out
echo Y | $MCM -P tests/fixfunctionality Fix.createFile -r $TEMP/fixfun_prompt_Y -f prompt | diffwithstdin tests/fixfunctionality/fix_prompt_Y.out
test -e $TEMP/fixfun_prompt_Y/fixfunctionality || (echo "Error: fix_prompt_Y file not created"; exit 1)
$MCM -P tests/fixfunctionality Fix.createFile -r $TEMP/fixfun_ifnew -f ifnew | diffwithstdin tests/fixfunctionality/fix_ifnew.out
$MCM -P tests/fixfunctionality Fix.createFile -r $TEMP -f ifnew | diffwithstdin tests/fixfunctionality/fix_ifnew_alreadyexists.out
$MCM -P tests/fixfunctionality Fix.createFile -r $TEMP/fixfun_promptunless -f promptunlessonlyadditions | diffwithstdin tests/fixfunctionality/fix_promptunless.out
$MCM -P tests/fixfunctionality Fix.createFile -r $TEMP -f promptunlessonlyadditions | diffwithstdin tests/fixfunctionality/fix_promptunless_alreadyexists.out
rm $TEMP/fixfunctionality
mkdir $TEMP/fixfunctionality
echo no | $MCM -P tests/fixfunctionality Fix.createFile -r $TEMP -f promptunlessonlyadditions | diffwithstdin tests/fixfunctionality/fix_promptunless_requireschanges.out

echo
echo "Checking generation of compiled files"
set -e
test ! -e $TEMP/cout || (echo "Error: Impossible!  $TEMP/cout already exists"; exit 1)
for i in tests/*.mcm;do
    echo $i
    expected=`echo $i|sed s/mcm/compiled/g`
    $MCM Test.test -P tests -f ifnew -r $TEMP/mcm_tests --stdin -c $TEMP/cout <$i >/dev/null
    cat $TEMP/cout/Test_test.mcm | diffwithstdin $expected
    $MCM Test_test.test -f no -r / --stdin -d <$TEMP/cout/Test_test.mcm
    rm $TEMP/cout/Test_test.mcm
    rmdir $TEMP/cout
done
$MCM -P tests -f ifnew -r / --stdin -c $TEMP/cout <tests/Fragment1.mcm Test.test Test.test1 Test.test2 Test.test3 Test.test4 Test.test5 Test.test6 | diffwithstdin tests/Fragment1.manycompiledoutput
rm $TEMP/cout/Test_test1.mcm $TEMP/cout/Test_test4.mcm $TEMP/cout/Test_test.mcm $TEMP/cout/Test_test2.mcm $TEMP/cout/Test_test5.mcm $TEMP/cout/Test_test3.mcm $TEMP/cout/Test_test6.mcm
rmdir $TEMP/cout

echo
echo "Checking status/pathtype matrix"
set -e
STATUSMCM="$MCM -P teststatuspathtype -r $TEMP/mcm_tests/statuspathtype -d"
DOTSTATUSES=".missing .wrongtype .wrongcontents .extrafiles .wrongpermissions .perfect"
PLAINSTATUSES=`echo $DOTSTATUSES | sed 's/\.//g'`
$STATUSMCM -f yes .init | diffwithstdin teststatuspathtype/init.expected
for i in $PLAINSTATUSES;do
    $STATUSMCM -f no .$i | diffwithstdin teststatuspathtype/$i.expected
done

echo
echo "Checking parsing of multiple startingpoints"
set -e
$STATUSMCM --parse-only --error-on-leftover-fragments .init $DOTSTATUSES

echo
echo "Simplistic mcm2html tests"
echo
mkdir $TEMP/mcm2htmltests
cp -a tests/*.{mcm,compiled} $TEMP/mcm2htmltests/
find $TEMP/mcm2htmltests \( -name '*.mcm' -o -name '*.compiled' \) -type f -exec $MCM2HTML '{}' ';' >/dev/null
for i in `ls testhtmlresults`;do
    diff testhtmlresults/$i $TEMP/mcm2htmltests/$i || (echo "Diff failed for $i"; exit 1)
done

echo
echo "Simplistic commands2html tests"
echo
mkdir $TEMP/commands2htmltests
cp -a tests/commands2html/* $TEMP/commands2htmltests/
for i in $TEMP/commands2htmltests/*.commands;do
    cd $TEMP/commands2htmltests
    $COMMANDS2HTML $i
    cd -
    o=tests/commands2html/`basename $i`.html_snippet_expected
    diff $i.html_snippet $o || (echo "Diff failed for $i"; exit 1)
done
